
/*
    fft.c - fast fourier transform using libfftw3
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

/* 
    VE3NEA:

    Gain := (1 - Exp(-(0.2 * NewSpectrum[i])));
    FilteredSpectrum[i] := FilteredSpectrum[i] * (1-Gain) + NewSpectrum[i] * Gain;
*/


struct fft *fft;

struct fft *init_fft(void){
    struct fft *fft = NULL;
	int i,j,r,f;

#ifdef USE_FFT
    if (!sdl) return NULL;

    fft = g_new0(struct fft, 1);
    fft->n = -1;
    fft->rin = NULL;
    
	j=0;
	for (i = 128; i < 256; i++){
		r = (i-128)*2;
		f = 255 - r;
		fft->pal[j++] = makecol(0, 0, r);
	} 
	for (i = 256; i < 512; i++){
		r = i - 256;
		f = 255 - r;
		fft->pal[j++] = makecol(0, r, 255);
	}
	for (i = 512; i < 768; i++){
		r = i - 512;
		f = 255 - r;
		fft->pal[j++] = makecol(r, 255, f);
	}
	for (i = 768; i < 1024-32; i++){
		r = i - 768;
		f = 255 - r;
		fft->pal[j++] = makecol(255, f, 0);
	}
	if (j != FFT_COLORS) {
		//dbg("j=%d\n", j);
		internal("bad fft color number, please fix");
	}
    MUTEX_INIT(fft->screen);

#endif
    return fft;
}

void free_fft(struct fft *fft){
    if (!fft) return;
#ifdef USE_FFT
    dbg("free_fft()\n");

    fftw_destroy_plan(fft->plan);
    fftw_free(fft->rin);
    fftw_free(fft->cout);

    if (fft->screen) SDL_FreeSurface(fft->screen);
    MUTEX_FREE(fft->screen);

    g_free(fft);
#endif
}

void fft_start(int n){
    int i;
#ifdef USE_FFT
    if (!fft) return;
    //dbg("fft_start(%d) ch=%d\n", n, dsp->channels);
	if (n != fft->n){ 
     //   dbg("    realloc %d->%d\n", fft->n, n);
		fft->n = n;
		if (fft->rin){
			fftw_destroy_plan(fft->plan);
			fftw_free(fft->rin);
			fftw_free(fft->cout);
		}
		fft->rin = (double *) fftw_malloc(sizeof(double) * fft->n + 256); // fftw_malloc seems not to add padding elements
		//fft->cout = (double *) fftw_malloc(sizeof(double) * fft->n + 256);
		fft->cout = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * fft->n + 256);
    	fft->plan = fftw_plan_dft_r2c_1d(fft->n, fft->rin,  fft->cout, FFTW_ESTIMATE);
	}

    fft->screeny = 0;
    memset(fft->ma_amp, 0, sizeof(fft->ma_amp));
    for (i=0; i<FFT_AMPLEN; i++) fft->nea_amp[i] = 0.0;

#endif
}

void fft_resize(struct fft *fft, struct subwin *sw){
#ifdef USE_FFT
    fft->wf_x2 = 260;
    fft->wf_x3 = 2*260;

    sw->fft_sp_h = 210;
    sw->fft_sp_y = sw->h*FONT_H - sw->fft_sp_h;
    sw->fft_wf_y = 0;
    sw->fft_wf_h = sw->fft_sp_y;
    sw->fft_sp_x2 = fft->wf_x2;
    sw->fft_wf_x2 = fft->wf_x2;

    SDL_Surface *newscr;
    newscr = SDL_CreateRGBSurface(SDL_SWSURFACE, sw->w*FONT_W , sw->h*FONT_H, sdl->bpp, 
        sdl->screen->format->Rmask, sdl->screen->format->Gmask, sdl->screen->format->Gmask, 0);
    MUTEX_LOCK(fft->screen);
    if (fft->screen){
        //SDL_BlitSurface(fft->screen, NULL, newscr, NULL);
        SDL_FreeSurface(fft->screen);
    }
    fft->screen = newscr;
    fft->screeny = 0;
    MUTEX_UNLOCK(fft->screen);
#endif
}

void sw_fft_redraw(struct subwin *sw){ 
#ifdef USE_FFT
    int i, y, sp_y, amp;
    SDL_Rect src, dst;

    MUTEX_LOCK(fft->screen);
    if (fft->screeny <= 0){
        MUTEX_UNLOCK(fft->screen);
        return;
    }
    y = fft->screeny;
    fft->screeny = 0;


    MUTEX_LOCK(sw->screen);

    // move old screen
    src.x = 0;
    src.y = sw->fft_wf_y + y;
    src.w = sw->screen->w;
    src.h = sw->fft_wf_h - y;
    dst.x = 0;
    dst.y = sw->fft_wf_y;
    dst.w = sw->screen->w;
    dst.h = sw->fft_wf_h - y;
    SDL_BlitSurface(sw->screen, &src, sw->screen, &dst);
    
    // add new data
    src.x = 0;
    src.y = 0;
    src.w = fft->screen->w;
    src.h = y;
    dst.x = 0;
    dst.y = sw->fft_wf_y + sw->fft_wf_h - y;
    dst.w = sw->screen->w;
    dst.h = y;
    SDL_BlitSurface(fft->screen, &src, sw->screen, &dst);

    // clear background under spectral lines
    dst.x = 0;
    dst.w = sw->screen->w;
    dst.y = sw->fft_sp_y;
    dst.h = sw->fft_sp_h;
    SDL_FillRect(sw->screen, &dst, 0);

    // draw spetral lines
    sp_y = sw->fft_sp_y + sw->fft_sp_h - 1;
    for (i=0; i<=FFT_AMPLEN; i++){
        amp = fft->amp[i] * (sw->fft_sp_h - 3);
        line(sw->screen, i, sp_y, i, sp_y - amp, sdl->gr[12]);

        amp = fft->ma_sum[i] * (sw->fft_sp_h - 3);
        line(sw->screen, i+fft->wf_x2, sp_y, i+fft->wf_x2, sp_y - amp, sdl->gr[12]);

        amp = fft->nea_amp[i] * (sw->fft_sp_h - 3);
        line(sw->screen, i+fft->wf_x3, sp_y, i+fft->wf_x3, sp_y - amp, sdl->gr[12]);
    }
    MUTEX_UNLOCK(sw->screen);
    MUTEX_UNLOCK(fft->screen);
/*    for (i=0; i<FFT_COLORS; i++){
        line(sw->screen, i, 0, i, 10, fft->pal[i]);
    }*/
#endif
}

//int xxi=0x100;

void fft_do(struct fft *fft){
#ifdef USE_FFT
    int i, col/*, mini, maxi*/;
    double d, e;
//    double mind, maxd, mine, maxe, minv, maxv;
    int j;
    double gain;
/*    mini = 200000000;
    maxi = -200000000;
    mind = 1e30;
    maxd = -1e30;
    mine = 1e30;
    maxe = -1e30;
    minv = 1e30;
    maxv = -1e30;   */
#if 0
    for (i=0; i<fft->n; i++){
        fft->rin[i]=1.0 * cos(i* 1000 *2*MY_PI/8000.0) +
                    0.0 * cos(i* 500 *2*MY_PI/8000.0) +
                    0.0 * cos(i* 250 *2*MY_PI/8000.0) +
                    0.0 * cos(i* 125 *2*MY_PI/8000.0) +
                    0;
       
        d = fft->rin[i];
    }
#endif  
    if (!fft->screen) return;
    
    fftw_execute(fft->plan);

//    dbg("%f\t%f\t%f\t%f\t%f\t%f\n", fft->cout[0], fft->cout[1], fft->cout[2], fft->cout[3], fft->cout[4], fft->cout[5]);
/*    for (i=0; i<6;i++){
        dbg("fft->cout=%f+%fi\t", fft->cout[i][0], fft->cout[i][1]);
    }
    dbg("\n");
    internal(""); */

/*    for (i=0; i<fft->n; i++){
        fft->cout[i]=256;
    }*/

    MUTEX_LOCK(fft->screen);
    for (i=0; i<=fft->n/2; i++){
//    for (i=60; i<=80; i++){
        if (i > fft->screen->h) break;
                        
        d = fabs(sqrt(sqr(fft->cout[i][0])+sqr(fft->cout[i][1])));
        d = d/(fft->n/2);  
        // d is epsilon to 1.0
        e = log10(d);
//        printf("%3d: %f\t%f\n", i, d, e);
        e -= log10(1.0/32768.0);
//        printf("e=%f %f\n", e, (log10(1.0)-log(1.0/32768.0)));
        e /= -log10(1.0/32768.0);
        if (e<0) e = 0;  // slabsi signaly nez 1 bit
        //e /= -log10(1.0/32768.0);
//        col = (int)e * FFT_COLORS;

/*        if (d<mind) mind = d;
        if (d>maxd) maxd = d;
        if (d<mine) mine = e;
        if (d>maxe) maxe = e;
        if (col<mini) mini=col;
        if (col>maxi) maxi=col;

        if (col < 0) col = 0;
        if (col >= FFT_COLORS) col = FFT_COLORS - 1;*/
//            col = xxi;
        fft->amp[i] = e;
        col = (int)(e * (FFT_COLORS - 1.0));
//        dbg("d=%11.9f e=%f\n", d, e);
//        dbg("col=%d\n");
        fast_putpixel(fft->screen, i, fft->screeny, fft->pal[col]);

        fft->ma_amp[fft->ma_i][i] = e;
        fft->ma_sum[i] = 0;
        for (j=0; j<FFT_MA; j++) fft->ma_sum[i] += fft->ma_amp[j][i];
        fft->ma_sum[i] /= FFT_MA;
        col = fft->ma_sum[i] * (FFT_COLORS - 1);
        fast_putpixel(fft->screen, i+fft->wf_x2, fft->screeny, fft->pal[col]);

/*        gain = fft->nea_amp[i];
        gain *= (-log10(1.0/32738.0));
        gain += log10(1.0/32768.0);
        gain *= log(10.0);
        gain = exp(gain);
        if (fft->nea_first) gain = 1;
        fft->nea_first = 0;          */
        gain = d;
        if (gain < 0.03) gain = 0.03;

//        dbg("gain=%f\n", gain);
        fft->nea_amp[i] = fft->nea_amp[i] * (1-gain) + e * gain;
        fast_putpixel(fft->screen, i+fft->wf_x3, fft->screeny, fft->pal[(int)(fft->nea_amp[i]*(FFT_COLORS-3))]);
    }    
/*       xxi++;
       xxi %= FFT_COLORS;*/
    fft->screeny++;
    fft->ma_i++;
    fft->ma_i %= FFT_MA;
    if (fft->screeny == fft->screen->h) fft->screeny = 0;
    MUTEX_UNLOCK(fft->screen);
//    dbg("v=%f .. %f   \td=%f .. %f     \te=%f .. %f     \tcol=%d .. %d\n", minv, maxv, mind, maxd, mine, maxe, mini, maxi);
//    internal("");
#endif
}
